# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>

# for internal use only
from __future__ import annotations

import argparse
import re
import sys

import rich_argparse._lazy_rich as r
from rich_argparse._common import (
    _HIGHLIGHTS,
    _fix_legacy_win_text,
    _strip_codes,
    rich_fill,
    rich_strip,
    rich_wrap,
)

TYPE_CHECKING = False
if TYPE_CHECKING:
    from argparse import Action, ArgumentParser, Namespace, _MutuallyExclusiveGroup
    from collections.abc import Callable, Iterable, Iterator, MutableMapping, Sequence
    from typing import Any, ClassVar
    from typing_extensions import Self


class RichHelpFormatter(argparse.HelpFormatter):
    """An argparse HelpFormatter class that renders using rich."""

    group_name_formatter: ClassVar[Callable[[str], str]] = str.title
    """A function that formats group names. Defaults to ``str.title``."""
    styles: ClassVar[dict[str, r.StyleType]] = {
        "argparse.args": "cyan",
        "argparse.groups": "dark_orange",
        "argparse.help": "default",
        "argparse.metavar": "dark_cyan",
        "argparse.syntax": "bold",
        "argparse.text": "default",
        "argparse.prog": "grey50",
        "argparse.default": "italic",
    }
    """A dict of rich styles to control the formatter styles.

    The following styles are used:

    - ``argparse.args``: for positional-arguments and --options (e.g "--help")
    - ``argparse.groups``: for group names (e.g. "positional arguments")
    - ``argparse.help``: for argument's help text (e.g. "show this help message and exit")
    - ``argparse.metavar``: for meta variables (e.g. "FILE" in "--file FILE")
    - ``argparse.prog``: for %(prog)s in the usage (e.g. "foo" in "Usage: foo [options]")
    - ``argparse.syntax``: for highlights of back-tick quoted text (e.g. "``` `some text` ```")
    - ``argparse.text``: for the descriptions and epilog (e.g. "A foo program")
    - ``argparse.default``: for %(default)s in the help (e.g. "Value" in "(default: Value)")
    """
    highlights: ClassVar[list[str]] = _HIGHLIGHTS[:]
    """A list of regex patterns to highlight in the help text.

    It is used in the description, epilog, groups descriptions, and arguments' help. By default,
    it highlights ``--words-with-dashes`` with the `argparse.args` style and
    `` `text in backquotes` `` with the `argparse.syntax` style.

    To disable highlighting, clear this list (``RichHelpFormatter.highlights.clear()``).
    """
    usage_markup: ClassVar[bool] = False
    """If True, render the usage string passed to ``ArgumentParser(usage=...)`` as markup.

    Defaults to ``False`` meaning the text of the usage will be printed verbatim.

    Note that the auto-generated usage string is always colored.
    """
    help_markup: ClassVar[bool] = True
    """If True (default), render the help message of arguments as console markup."""
    text_markup: ClassVar[bool] = True
    """If True (default), render the descriptions and epilog as console markup."""

    _root_section: _Section
    _current_section: _Section

    def __init__(
        self,
        prog: str,
        indent_increment: int = 2,
        max_help_position: int = 24,
        width: int | None = None,
        *,
        console: r.Console | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(prog, indent_increment, max_help_position, width, **kwargs)
        self._console = console

        # https://docs.python.org/3/library/stdtypes.html#printf-style-string-formatting
        self._printf_style_pattern = re.compile(
            r"""
            %                               # Percent character
            (?:\((?P<mapping>[^)]*)\))?     # Mapping key
            (?P<flag>[#0\-+ ])?             # Conversion Flags
            (?P<width>\*|\d+)?              # Minimum field width
            (?P<precision>\.(?:\*?|\d*))?   # Precision
            [hlL]?                          # Length modifier (ignored)
            (?P<format>[diouxXeEfFgGcrsa%]) # Conversion type
            """,
            re.VERBOSE,
        )

    @property
    def console(self) -> r.Console:
        if self._console is None:
            self._console = r.Console()
        return self._console

    @console.setter
    def console(self, console: r.Console) -> None:
        self._console = console

    if sys.version_info >= (3, 14):  # pragma: >=3.14 cover

        def _set_color(self, color: bool, *args, **kwargs) -> None:
            # Override to disable color setting in argparse.HelpFormatter for Python 3.14+
            return super()._set_color(False, *args, **kwargs)

    class _Section(argparse.HelpFormatter._Section):
        def __init__(
            self, formatter: RichHelpFormatter, parent: Self | None, heading: str | None = None
        ) -> None:
            if heading is not argparse.SUPPRESS and heading is not None:
                heading = f"{type(formatter).group_name_formatter(heading)}:"
            super().__init__(formatter, parent, heading)
            self.formatter: RichHelpFormatter
            self.rich_items: list[r.RenderableType] = []
            self.rich_actions: list[tuple[r.Text, r.Text | None]] = []
            if parent is not None:
                parent.rich_items.append(self)

        def _render_items(self, console: r.Console, options: r.ConsoleOptions) -> r.RenderResult:
            if not self.rich_items:
                return
            generated_options = options.update(no_wrap=True, overflow="ignore")
            new_line = r.Segment.line()
            for item in self.rich_items:
                if isinstance(item, RichHelpFormatter._Section):
                    yield from console.render(item, options)
                elif isinstance(item, r.Padding):  # user added rich renderable
                    yield from console.render(item, options)
                    yield new_line
                else:  # argparse generated rich renderable
                    yield from console.render(item, generated_options)

        def _render_actions(self, console: r.Console, options: r.ConsoleOptions) -> r.RenderResult:
            if not self.rich_actions:
                return
            options = options.update(no_wrap=True, overflow="ignore")
            help_pos = min(self.formatter._action_max_length + 2, self.formatter._max_help_position)
            help_width = max(self.formatter._width - help_pos, 11)
            indent = r.Text(" " * help_pos)
            for action_header, action_help in self.rich_actions:
                if not action_help:
                    # no help, yield the header and finish
                    yield from console.render(action_header, options)
                    continue
                action_help_lines = self.formatter._rich_split_lines(action_help, help_width)
                if len(action_header) > help_pos - 2:
                    # the header is too long, put it on its own line
                    yield from console.render(action_header, options)
                    action_header = indent
                action_header.set_length(help_pos)
                action_help_lines[0].rstrip()
                yield from console.render(action_header + action_help_lines[0], options)
                for line in action_help_lines[1:]:
                    line.rstrip()
                    yield from console.render(indent + line, options)
            yield ""

        def __rich_console__(self, console: r.Console, options: r.ConsoleOptions) -> r.RenderResult:
            if not self.rich_items and not self.rich_actions:
                return  # empty section
            if self.heading is not argparse.SUPPRESS and self.heading is not None:
                yield r.Text(self.heading, style="argparse.groups", overflow="ignore")
            yield from self._render_items(console, options)
            yield from self._render_actions(console, options)

    def __rich_console__(self, console: r.Console, options: r.ConsoleOptions) -> r.RenderResult:
        with console.use_theme(r.Theme(self.styles)):
            root = console.render(self._root_section, options.update_width(self._width))
            new_line = r.Segment.line()
            add_empty_line = False
            for line_segments in r.Segment.split_lines(root):
                for i, segment in enumerate(reversed(line_segments), start=1):
                    stripped = segment.text.rstrip()
                    if stripped:
                        if add_empty_line:
                            yield new_line
                        add_empty_line = False
                        yield from line_segments[:-i]
                        yield r.Segment(stripped, style=segment.style, control=segment.control)
                        yield new_line
                        break
                else:  # empty line
                    add_empty_line = True

    def add_text(self, text: r.RenderableType | None) -> None:
        if text is argparse.SUPPRESS or text is None:
            return
        elif isinstance(text, str):
            self._current_section.rich_items.append(self._rich_format_text(text))
        else:
            self.add_renderable(text)

    def add_renderable(self, renderable: r.RenderableType) -> None:
        padded = r.Padding.indent(renderable, self._current_indent)
        self._current_section.rich_items.append(padded)

    def add_usage(
        self,
        usage: str | None,
        actions: Iterable[Action],
        groups: Iterable[_MutuallyExclusiveGroup],
        prefix: str | None = None,
    ) -> None:
        if usage is argparse.SUPPRESS:
            return
        if prefix is None:
            prefix = self._format_usage(usage="", actions=(), groups=(), prefix=None).rstrip("\n")
        prefix = _strip_codes(prefix)
        prefix_end = ": " if prefix.endswith(": ") else ""
        prefix = prefix[: len(prefix) - len(prefix_end)]
        prefix = type(self).group_name_formatter(prefix) + prefix_end

        usage_spans = [r.Span(0, len(prefix.rstrip()), "argparse.groups")]
        usage_text = _strip_codes(self._format_usage(usage, actions, groups, prefix=prefix))
        if usage is None:  # get colour spans for generated usage
            prog = _strip_codes(f"{self._prog}")
            if actions:
                prog_start = usage_text.index(prog, len(prefix))
                usage_spans.append(r.Span(prog_start, prog_start + len(prog), "argparse.prog"))
            actions_start = len(prefix) + len(prog) + 1
            try:
                spans = list(self._rich_usage_spans(usage_text, actions_start, actions=actions))
            except ValueError:
                spans = []
            usage_spans.extend(spans)
            rich_usage = r.Text(usage_text)
        elif self.usage_markup:  # treat user provided usage as markup
            usage_spans.extend(self._rich_prog_spans(prefix + r.Text.from_markup(usage).plain))
            rich_usage = r.Text.from_markup(usage_text)
            usage_spans.extend(rich_usage.spans)
            rich_usage.spans.clear()
        else:  # treat user provided usage as plain text
            usage_spans.extend(self._rich_prog_spans(prefix + usage))
            rich_usage = r.Text(usage_text)
        rich_usage.spans.extend(usage_spans)
        self._root_section.rich_items.append(rich_usage)

    def add_argument(self, action: Action) -> None:
        super().add_argument(action)
        if action.help is not argparse.SUPPRESS:
            self._current_section.rich_actions.extend(self._rich_format_action(action))

    def format_help(self) -> str:
        with self.console.capture() as capture:
            self.console.print(self, crop=False)
        return _fix_legacy_win_text(self.console, capture.get())

    # ===============
    # Utility methods
    # ===============
    def _rich_prog_spans(self, usage: str) -> Iterator[r.Span]:
        if "%(prog)" not in usage:
            return
        params = {"prog": self._prog}
        formatted_usage = ""
        last = 0
        for m in self._printf_style_pattern.finditer(usage):
            start, end = m.span()
            formatted_usage += usage[last:start]
            sub = usage[start:end] % params
            prog_start = len(formatted_usage)
            prog_end = prog_start + len(sub)
            formatted_usage += sub
            last = end
            yield r.Span(prog_start, prog_end, "argparse.prog")

    def _rich_usage_spans(
        self, text: str, start: int, actions: Iterable[Action]
    ) -> Iterator[r.Span]:
        options: list[Action] = []
        positionals: list[Action] = []
        for action in actions:
            if action.help is not argparse.SUPPRESS:
                options.append(action) if action.option_strings else positionals.append(action)
        pos = start

        def find_span(_string: str) -> tuple[int, int]:
            stripped = _strip_codes(_string)
            _start = text.index(stripped, pos)
            _end = _start + len(stripped)
            return _start, _end

        for action in options:  # start with the options
            usage = action.format_usage()
            if isinstance(action, argparse.BooleanOptionalAction):
                for option_string in action.option_strings:
                    start, end = find_span(option_string)
                    yield r.Span(start, end, "argparse.args")
                    pos = end + 1
                continue
            start, end = find_span(usage)
            yield r.Span(start, end, "argparse.args")
            pos = end + 1
            if action.nargs != 0:
                default_metavar = self._get_default_metavar_for_optional(action)
                for metavar_part, colorize in self._rich_metavar_parts(action, default_metavar):
                    start, end = find_span(metavar_part)
                    if colorize:
                        yield r.Span(start, end, "argparse.metavar")
                    pos = end
            pos = end + 1
        for action in positionals:  # positionals come at the end
            default_metavar = self._get_default_metavar_for_positional(action)
            for metavar_part, colorize in self._rich_metavar_parts(action, default_metavar):
                start, end = find_span(metavar_part)
                if colorize:
                    yield r.Span(start, end, "argparse.args")
                pos = end
            pos = end + 1

    def _rich_metavar_parts(
        self, action: Action, default_metavar: str
    ) -> Iterator[tuple[str, bool]]:
        get_metavar = self._metavar_formatter(action, default_metavar)
        # similar to self._format_args but yields (part, colorize) of the metavar
        if action.nargs is None:
            # '%s' % get_metavar(1)
            yield "%s" % get_metavar(1), True  # noqa: UP031
        elif action.nargs == argparse.OPTIONAL:
            # '[%s]' % get_metavar(1)
            yield from (
                ("[", False),
                ("%s" % get_metavar(1), True),  # noqa: UP031
                ("]", False),
            )
        elif action.nargs == argparse.ZERO_OR_MORE:
            if len(get_metavar(1)) == 2:
                metavar = get_metavar(2)
                # '[%s [%s ...]]' % metavar
                yield from (
                    ("[", False),
                    ("%s" % metavar[0], True),  # noqa: UP031
                    (" [", False),
                    ("%s" % metavar[1], True),  # noqa: UP031
                    (" ", False),
                    ("...", True),
                    ("]]", False),
                )
            else:
                # '[%s ...]' % metavar
                yield from (
                    ("[", False),
                    ("%s" % get_metavar(1), True),  # noqa: UP031
                    (" ", False),
                    ("...", True),
                    ("]", False),
                )
        elif action.nargs == argparse.ONE_OR_MORE:
            # '%s [%s ...]' % get_metavar(2)
            metavar = get_metavar(2)
            yield from (
                ("%s" % metavar[0], True),  # noqa: UP031
                (" [", False),
                ("%s" % metavar[1], True),  # noqa: UP031
                (" ", False),
                ("...", True),
                ("]", False),
            )
        elif action.nargs == argparse.REMAINDER:
            # '...'
            yield "...", True
        elif action.nargs == argparse.PARSER:
            # '%s ...' % get_metavar(1)
            yield from (
                ("%s" % get_metavar(1), True),  # noqa: UP031
                (" ", False),
                ("...", True),
            )
        elif action.nargs == argparse.SUPPRESS:
            # ''
            yield "", False
        else:
            metavar = get_metavar(action.nargs)  # type: ignore[arg-type]
            first = True
            for met in metavar:
                if first:
                    first = False
                else:
                    yield " ", False
                yield "%s" % met, True  # noqa: UP031

    def _rich_whitespace_sub(self, text: r.Text) -> r.Text:
        # do this `self._whitespace_matcher.sub(' ', text).strip()` but text is Text
        spans = [m.span() for m in self._whitespace_matcher.finditer(text.plain)]
        for start, end in reversed(spans):
            if end - start > 1:  # slow path
                space = text[start : start + 1]
                space.plain = " "
                text = text[:start] + space + text[end:]
            else:  # performance shortcut
                text.plain = text.plain[:start] + " " + text.plain[end:]
        return rich_strip(text)

    # =====================================
    # Rich version of HelpFormatter methods
    # =====================================
    def _rich_expand_help(self, action: Action) -> r.Text:
        params = dict(vars(action), prog=self._prog)
        for name in list(params):
            if params[name] is argparse.SUPPRESS:
                del params[name]
            elif hasattr(params[name], "__name__"):
                params[name] = params[name].__name__
        if params.get("choices") is not None:
            params["choices"] = ", ".join([str(c) for c in params["choices"]])
        help_string = self._get_help_string(action)
        assert help_string is not None
        # raise ValueError if needed
        help_string % params  # pyright: ignore[reportUnusedExpression]
        parts = []
        defaults: list[str] = []
        default_sub_template = "rich-argparse-f3ae8b55df34d5d83a8189d2e4766e68-{}-argparse-rich"
        default_n = 0
        last = 0
        for m in self._printf_style_pattern.finditer(help_string):
            start, end = m.span()
            parts.append(help_string[last:start])
            sub = help_string[start:end] % params
            if m.group("mapping") == "default":
                defaults.append(sub)
                sub = default_sub_template.format(default_n)
                default_n += 1
            else:
                sub = r.escape(sub)
            parts.append(sub)
            last = end
        parts.append(help_string[last:])
        rich_help = (
            r.Text.from_markup("".join(parts), style="argparse.help")
            if self.help_markup
            else r.Text("".join(parts), style="argparse.help")
        )
        for i, default in reversed(list(enumerate(defaults))):
            default_sub = default_sub_template.format(i)
            try:
                start = rich_help.plain.rindex(default_sub)
            except ValueError:
                # This could happen in cases like `[default: %(default)s]` with markup activated
                import warnings

                action_id = next(iter(action.option_strings), action.dest)
                printf_pat = self._printf_style_pattern.pattern
                repl = next(
                    (
                        repr(m.group(1))[1:-1]
                        for m in re.finditer(rf"\[([^\]]*{printf_pat}[^\]]*)\]", help_string, re.X)
                        if m.group("mapping") == "default"
                    ),
                    "default: %(default)s",
                )
                msg = (
                    f"Failed to process default value in help string of argument {action_id!r}."
                    f"\nHint: try disabling rich markup: `RichHelpFormatter.help_markup = False`"
                    f"\n      or replace brackets by parenthesis: `[{repl}]` -> `({repl})`"
                )
                warnings.warn(msg, UserWarning, stacklevel=4)
                continue
            end = start + len(default_sub)
            rich_help = (
                rich_help[:start].append(default, style="argparse.default").append(rich_help[end:])
            )
        for highlight in self.highlights:
            rich_help.highlight_regex(highlight, style_prefix="argparse.")
        return rich_help

    def _rich_format_text(self, text: str) -> r.Text:
        if "%(prog)" in text:
            text = text % {"prog": r.escape(self._prog)}
        rich_text = (
            r.Text.from_markup(text, style="argparse.text")
            if self.text_markup
            else r.Text(text, style="argparse.text")
        )
        for highlight in self.highlights:
            rich_text.highlight_regex(highlight, style_prefix="argparse.")
        text_width = max(self._width - self._current_indent * 2, 11)
        indent = r.Text(" " * self._current_indent)
        return self._rich_fill_text(rich_text, text_width, indent)

    def _rich_format_action(self, action: Action) -> Iterator[tuple[r.Text, r.Text | None]]:
        header = self._rich_format_action_invocation(action)
        header.pad_left(self._current_indent)
        help = self._rich_expand_help(action) if action.help and action.help.strip() else None
        yield header, help
        for subaction in self._iter_indented_subactions(action):
            yield from self._rich_format_action(subaction)

    def _rich_format_action_invocation(self, action: Action) -> r.Text:
        if not action.option_strings:
            return r.Text().append(self._format_action_invocation(action), style="argparse.args")
        else:
            action_header = r.Text(", ").join(
                r.Text(o, "argparse.args") for o in action.option_strings
            )
            if action.nargs != 0:
                default = self._get_default_metavar_for_optional(action)
                action_header.append(" ")
                for metavar_part, colorize in self._rich_metavar_parts(action, default):
                    style = "argparse.metavar" if colorize else None
                    action_header.append(metavar_part, style=style)
            return action_header

    def _rich_split_lines(self, text: r.Text, width: int) -> r.Lines:
        return rich_wrap(self.console, self._rich_whitespace_sub(text), width)

    def _rich_fill_text(self, text: r.Text, width: int, indent: r.Text) -> r.Text:
        return rich_fill(self.console, self._rich_whitespace_sub(text), width, indent) + "\n\n"


class RawDescriptionRichHelpFormatter(RichHelpFormatter):
    """Rich help message formatter which retains any formatting in descriptions."""

    def _rich_fill_text(self, text: r.Text, width: int, indent: r.Text) -> r.Text:
        return r.Text("\n").join(indent + line for line in text.split()) + "\n\n"


class RawTextRichHelpFormatter(RawDescriptionRichHelpFormatter):
    """Rich help message formatter which retains formatting of all help text."""

    def _rich_split_lines(self, text: r.Text, width: int) -> r.Lines:
        return text.split()


class ArgumentDefaultsRichHelpFormatter(argparse.ArgumentDefaultsHelpFormatter, RichHelpFormatter):
    """Rich help message formatter which adds default values to argument help."""


class MetavarTypeRichHelpFormatter(argparse.MetavarTypeHelpFormatter, RichHelpFormatter):
    """Rich help message formatter which uses the argument 'type' as the default
    metavar value (instead of the argument 'dest').
    """


class HelpPreviewAction(argparse.Action):
    """Action that renders the help to SVG, HTML, or text file and exits."""

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str = argparse.SUPPRESS,
        default: str = argparse.SUPPRESS,
        help: str = argparse.SUPPRESS,
        *,
        path: str | None = None,
        export_kwds: MutableMapping[str, Any] | None = None,
    ) -> None:
        super().__init__(option_strings, dest, nargs="?", const=path, default=default, help=help)
        self.export_kwds = export_kwds or {}

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        path = values
        if path is None:
            parser.exit(1, "error: help preview path is not provided\n")
        if not isinstance(path, str):
            parser.exit(1, "error: help preview path must be a string\n")
        if not path.endswith((".svg", ".html", ".txt")):
            parser.exit(1, "error: help preview path must end with .svg, .html, or .txt\n")
        import io

        text = r.Text.from_ansi(parser.format_help())
        console = r.Console(file=io.StringIO(), record=True)
        console.print(text, crop=False)

        if path.endswith(".svg"):
            self.export_kwds.setdefault("title", "")
            console.save_svg(path, **self.export_kwds)
        elif path.endswith(".html"):
            console.save_html(path, **self.export_kwds)
        elif path.endswith(".txt"):
            console.save_text(path, **self.export_kwds)
        else:
            raise AssertionError("unreachable")
        parser.exit(0, f"Help preview saved to {path}\n")



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>
from __future__ import annotations

if __name__ == "__main__":
    import argparse
    import sys

    from rich.terminal_theme import DIMMED_MONOKAI

    from rich_argparse import HelpPreviewAction, RichHelpFormatter

    parser = argparse.ArgumentParser(
        prog="python -m rich_argparse",
        formatter_class=RichHelpFormatter,
        description=(
            "This is a [link https://pypi.org/project/rich]rich[/]-based formatter for "
            "[link https://docs.python.org/3/library/argparse.html#formatter-class]"
            "argparse's help output[/].\n\n"
            "It enables you to use the powers of rich like markup and highlights in your CLI help. "
        ),
        epilog=":link: Read more at https://github.com/hamdanal/rich-argparse#usage.",
    )
    parser.add_argument(
        "formatter-class",
        help=(
            "Simply pass `formatter_class=RichHelpFormatter` to the argument parser to get a "
            "colorful help like this."
        ),
    )
    parser.add_argument(
        "styles",
        help="Customize your CLI's help with the `RichHelpFormatter.styles` dictionary.",
    )
    parser.add_argument(
        "--highlights",
        metavar="REGEXES",
        help=(
            "Highlighting the help text is managed by the list of regular expressions "
            "`RichHelpFormatter.highlights`. Set to empty list to turn off highlighting.\n"
            "See the next two options for default values."
        ),
    )
    parser.add_argument(
        "--syntax",
        default=RichHelpFormatter.styles["argparse.syntax"],
        help=(
            "Text inside backticks is highlighted using the `argparse.syntax` style "
            "(default: %(default)r)"
        ),
    )
    parser.add_argument(
        "-o",
        "--option",
        metavar="METAVAR",
        help="Text that looks like an --option is highlighted using the `argparse.args` style.",
    )
    group = parser.add_argument_group(
        "more arguments",
        description=(
            "This is a custom group. Group names are [italic]*Title Cased*[/] by default. Use the "
            "`RichHelpFormatter.group_name_formatter` function to change their format."
        ),
    )
    group.add_argument(
        "--more",
        nargs="*",
        help="This formatter works with subparsers, mutually exclusive groups and hidden arguments.",
    )
    mutex = group.add_mutually_exclusive_group()
    mutex.add_argument(
        "--rich",
        action="store_true",
        help="Rich and poor are mutually exclusive. Choose either one but not both.",
    )
    mutex.add_argument(
        "--poor", action="store_false", dest="rich", help="Does poor mean --not-rich 😉?"
    )
    mutex.add_argument("--not-rich", action="store_false", dest="rich", help=argparse.SUPPRESS)
    parser.add_argument(
        "--generate-rich-argparse-preview",
        action=HelpPreviewAction,
        path="rich-argparse.svg",
        export_kwds={"theme": DIMMED_MONOKAI},
    )
    # There is no program to run, always print help (except for the hidden --generate option)
    # You probably don't want to do this in your own code.
    if any(arg.startswith("--generate") for arg in sys.argv):
        parser.parse_args()
    else:
        parser.print_help()



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>
"""Extra formatters for rich help messages.

The rich_argparse.contrib module contains optional, standard implementations of common patterns of
rich help message formatting. These formatters are not included in the main rich_argparse module
because they do not translate directly to argparse formatters.
"""

from __future__ import annotations

from rich_argparse._contrib import ExtendedParagraphRichHelpFormatter, ParagraphRichHelpFormatter

__all__ = [
    "ExtendedParagraphRichHelpFormatter",
    "ParagraphRichHelpFormatter",
]



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>
from __future__ import annotations

from rich_argparse._optparse import (
    GENERATE_USAGE,
    IndentedRichHelpFormatter,
    RichHelpFormatter,
    TitledRichHelpFormatter,
)

__all__ = [
    "RichHelpFormatter",
    "IndentedRichHelpFormatter",
    "TitledRichHelpFormatter",
    "GENERATE_USAGE",
]


if __name__ == "__main__":
    import optparse

    IndentedRichHelpFormatter.highlights.append(r"(?P<metavar>\bregexes\b)")
    parser = optparse.OptionParser(
        description="I [link https://pypi.org/project/rich]rich[/]ify:trade_mark: optparse help.",
        formatter=IndentedRichHelpFormatter(),
        prog="python -m rich_arparse.optparse",
        epilog=":link: https://github.com/hamdanal/rich-argparse#optparse-support.",
        usage=GENERATE_USAGE,
    )
    parser.add_option("--formatter", metavar="rich", help="A piece of :cake: isn't it? :wink:")
    parser.add_option(
        "--styles", metavar="yours", help="Not your style? No biggie, change it :sunglasses:"
    )
    parser.add_option(
        "--highlights",
        action="store_true",
        help=":clap: --highlight :clap: all :clap: the :clap: regexes :clap:",
    )
    parser.add_option(
        "--syntax", action="store_true", help="`backquotes` may be bold, but they are :muscle:"
    )
    parser.add_option(
        "-s", "--long", metavar="METAVAR", help="That's a lot of metavars for an option!"
    )

    group = parser.add_option_group("Magic", description=":sparkles: :sparkles: :sparkles:")
    group.add_option(
        "--treasure", action="store_false", help="Mmm, did you find the --hidden :gem:?"
    )
    group.add_option("--hidden", action="store_false", dest="treasure", help=optparse.SUPPRESS_HELP)
    parser.print_help()



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>

# for internal use only
from __future__ import annotations

import rich_argparse._lazy_rich as r
from rich_argparse._argparse import RichHelpFormatter
from rich_argparse._common import rich_strip, rich_wrap


class ParagraphRichHelpFormatter(RichHelpFormatter):
    """Rich help message formatter which retains paragraph separation with spacing using `\\n\\n`."""

    def _rich_split_lines(self, text: r.Text, width: int) -> r.Lines:
        text = rich_strip(text)
        lines = r.Lines()
        for paragraph in text.split("\n\n"):
            # Normalize whitespace in the paragraph
            paragraph = self._rich_whitespace_sub(paragraph)
            # Wrap the paragraph to the specified width
            paragraph_lines = rich_wrap(self.console, paragraph, width)
            # Add the wrapped lines to the output
            lines.extend(paragraph_lines)
            # Add a blank line between paragraphs
            lines.append(r.Text("\n"))
        if lines:  # pragma: no cover
            lines.pop()  # Remove trailing newline
        return lines

    def _rich_fill_text(self, text: r.Text, width: int, indent: r.Text) -> r.Text:
        lines = self._rich_split_lines(text, width)
        return r.Text("\n").join(indent + line for line in lines) + "\n"


class ExtendedParagraphRichHelpFormatter(RichHelpFormatter):
    """Rich help message formatter which retains paragraph separation without spacing using `\\n\\n` and
    paragraph spacing using `\\n\\n\\n`."""

    def _rich_split_lines(self, text: r.Text, width: int) -> r.Lines:
        text = rich_strip(text)
        lines = r.Lines()
        for paragraph in text.split("\n\n\n"):
            for subparagraph in paragraph.split("\n\n"):
                # Normalize whitespace in the subparagraph
                subparagraph = self._rich_whitespace_sub(subparagraph)
                # Wrap the subparagraph to the specified width
                paragraph_lines = rich_wrap(self.console, subparagraph, width)
                # Add the wrapped lines to the output
                lines.extend(paragraph_lines)
            # Add a blank line between paragraphs
            lines.append(r.Text("\n"))
        if lines:  # pragma: no cover
            lines.pop()  # Remove trailing newline
        return lines

    def _rich_fill_text(self, text: r.Text, width: int, indent: r.Text) -> r.Text:
        lines = self._rich_split_lines(text, width)
        return r.Text("\n").join(indent + line for line in lines) + "\n"



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>

# for internal use only
from __future__ import annotations

import sys

import rich_argparse._lazy_rich as r

# Default highlight patterns:
# - highlight `text in backquotes` as "syntax"
# - --words-with-dashes outside backticks as "args"
_HIGHLIGHTS = [
    r"`(?P<syntax>[^`]*)`|(?:^|\s)(?P<args>-{1,2}[\w]+[\w-]*)",
]

_windows_console_fixed: bool | None = None


def rich_strip(text: r.Text) -> r.Text:
    """Strip leading and trailing whitespace from `rich.text.Text`."""
    lstrip_at = len(text.plain) - len(text.plain.lstrip())
    if lstrip_at:  # rich.Text.lstrip() is not available yet!!
        text = text[lstrip_at:]
    text.rstrip()
    return text


def rich_wrap(console: r.Console, text: r.Text, width: int) -> r.Lines:
    """`textwrap.wrap()` equivalent for `rich.text.Text`."""
    text = text.copy()
    text.expand_tabs(8)  # textwrap expands tabs first
    whitespace_trans = dict.fromkeys(map(ord, "\t\n\x0b\x0c\r "), ord(" "))
    text.plain = text.plain.translate(whitespace_trans)
    return text.wrap(console, width)


def rich_fill(console: r.Console, text: r.Text, width: int, indent: r.Text) -> r.Text:
    """`textwrap.fill()` equivalent for `rich.text.Text`."""
    lines = rich_wrap(console, text, width)
    return r.Text("\n").join(indent + line for line in lines)


def _strip_codes(text: str) -> str:
    """Remove ANSI color codes and control codes from a string."""
    return r.re_ansi.sub("", r.strip_control_codes(text))


def _initialize_win_colors() -> bool:  # pragma: no cover
    global _windows_console_fixed
    assert sys.platform == "win32"
    if _windows_console_fixed is None:
        winver = sys.getwindowsversion()  # type: ignore[attr-defined]
        if winver.major < 10 or winver.build < 10586:
            try:
                import colorama

                _windows_console_fixed = isinstance(sys.stdout, colorama.ansitowin32.StreamWrapper)
            except Exception:
                _windows_console_fixed = False
        else:
            import ctypes

            kernel32 = ctypes.windll.kernel32  # type: ignore[attr-defined]
            ENABLE_PROCESSED_OUTPUT = 0x1
            ENABLE_WRAP_AT_EOL_OUTPUT = 0x2
            ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x4
            STD_OUTPUT_HANDLE = -11
            kernel32.SetConsoleMode(
                kernel32.GetStdHandle(STD_OUTPUT_HANDLE),
                ENABLE_PROCESSED_OUTPUT
                | ENABLE_WRAP_AT_EOL_OUTPUT
                | ENABLE_VIRTUAL_TERMINAL_PROCESSING,
            )
            _windows_console_fixed = True
    return _windows_console_fixed


def _fix_legacy_win_text(console: r.Console, text: str) -> str:
    # activate legacy Windows console colors if needed (and available) or strip ANSI escape codes
    if (
        text
        and sys.platform == "win32"
        and console.legacy_windows
        and console.color_system is not None
        and not _initialize_win_colors()
    ):  # pragma: win32 cover
        text = "\n".join(r.re_ansi.sub("", line) for line in text.split("\n"))
    return text



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>

# for internal use only
from __future__ import annotations

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any

    from rich.ansi import re_ansi as re_ansi
    from rich.console import Console as Console
    from rich.console import ConsoleOptions as ConsoleOptions
    from rich.console import RenderableType as RenderableType
    from rich.console import RenderResult as RenderResult
    from rich.containers import Lines as Lines
    from rich.control import strip_control_codes as strip_control_codes
    from rich.markup import escape as escape
    from rich.padding import Padding as Padding
    from rich.segment import Segment as Segment
    from rich.style import StyleType as StyleType
    from rich.text import Span as Span
    from rich.text import Text as Text
    from rich.theme import Theme as Theme

__all__ = [
    "re_ansi",
    "Console",
    "ConsoleOptions",
    "RenderableType",
    "RenderResult",
    "Lines",
    "strip_control_codes",
    "escape",
    "Padding",
    "Segment",
    "StyleType",
    "Span",
    "Text",
    "Theme",
]


def __getattr__(name: str) -> Any:
    if name not in __all__:
        raise AttributeError(name)
    import rich.ansi
    import rich.console
    import rich.containers
    import rich.control
    import rich.markup
    import rich.padding
    import rich.segment
    import rich.style
    import rich.text
    import rich.theme

    globals().update(
        {
            "re_ansi": rich.ansi.re_ansi,
            "Console": rich.console.Console,
            "ConsoleOptions": rich.console.ConsoleOptions,
            "RenderableType": rich.console.RenderableType,
            "RenderResult": rich.console.RenderResult,
            "Lines": rich.containers.Lines,
            "strip_control_codes": rich.control.strip_control_codes,
            "escape": rich.markup.escape,
            "Padding": rich.padding.Padding,
            "Segment": rich.segment.Segment,
            "StyleType": rich.style.StyleType,
            "Span": rich.text.Span,
            "Text": rich.text.Text,
            "Theme": rich.theme.Theme,
        }
    )
    return globals()[name]



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>

# for internal use only
from argparse import _FormatterClass
from collections.abc import Callable
from typing import TypeVar, overload

from rich_argparse._argparse import RichHelpFormatter

_T = TypeVar("_T", bound=type)

@overload
def patch_default_formatter_class(
    cls: None = None,
    /,
    *,
    formatter_class: _FormatterClass = RichHelpFormatter,
    method_name: str = "__init__",
) -> Callable[[_T], _T]: ...
@overload
def patch_default_formatter_class(
    cls: _T,
    /,
    *,
    formatter_class: _FormatterClass = RichHelpFormatter,
    method_name: str = "__init__",
) -> _T: ...



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>
from __future__ import annotations

from rich_argparse._argparse import (
    ArgumentDefaultsRichHelpFormatter,
    HelpPreviewAction,
    MetavarTypeRichHelpFormatter,
    RawDescriptionRichHelpFormatter,
    RawTextRichHelpFormatter,
    RichHelpFormatter,
)

__all__ = [
    "RichHelpFormatter",
    "RawDescriptionRichHelpFormatter",
    "RawTextRichHelpFormatter",
    "ArgumentDefaultsRichHelpFormatter",
    "MetavarTypeRichHelpFormatter",
    "HelpPreviewAction",
]



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>

# for internal use only
from __future__ import annotations

from rich_argparse._argparse import RichHelpFormatter


def patch_default_formatter_class(
    cls=None, /, *, formatter_class=RichHelpFormatter, method_name="__init__"
):
    """Patch the default `formatter_class` parameter of an argument parser constructor.

    Parameters
    ----------
    cls : (type, optional)
        The class to patch. If not provided, a decorator is returned.
    formatter_class : (type, optional)
        The new formatter class to use. Defaults to ``RichHelpFormatter``.
    method_name : (str, optional)
        The method name to patch. Defaults to ``__init__``.

    Examples
    --------
    Can be used as a normal function to patch an existing class::

        # Patch the default formatter class of `argparse.ArgumentParser`
        patch_default_formatter_class(argparse.ArgumentParser)

        # Patch the default formatter class of django commands
        from django.core.management.base import BaseCommand, DjangoHelpFormatter
        class DjangoRichHelpFormatter(DjangoHelpFormatter, RichHelpFormatter): ...
        patch_default_formatter_class(
            BaseCommand, formatter_class=DjangoRichHelpFormatter, method_name="create_parser"
        )
    Or as a decorator to patch a new class::

        @patch_default_formatter_class
        class MyArgumentParser(argparse.ArgumentParser):
            pass

        @patch_default_formatter_class(formatter_class=RawDescriptionRichHelpFormatter)
        class MyOtherArgumentParser(argparse.ArgumentParser):
            pass
    """
    import functools

    def decorator(cls, /):
        method = getattr(cls, method_name)
        if not callable(method):
            raise TypeError(f"'{cls.__name__}.{method_name}' is not callable")

        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            kwargs.setdefault("formatter_class", formatter_class)
            return method(*args, **kwargs)

        setattr(cls, method_name, wrapper)
        return cls

    if cls is None:
        return decorator
    return decorator(cls)



# Source code: https://github.com/hamdanal/rich-argparse
# MIT license: Copyright (c) Ali Hamdan <ali.hamdan.dev@gmail.com>
"""Django-specific utilities for rich command line help."""

from __future__ import annotations

try:
    from django.core.management.base import DjangoHelpFormatter as _DjangoHelpFormatter
except ImportError as e:  # pragma: no cover
    raise ImportError("rich_argparse.django requires django to be installed.") from e

from rich_argparse._argparse import RichHelpFormatter as _RichHelpFormatter
from rich_argparse._patching import patch_default_formatter_class as _patch_default_formatter_class

__all__ = [
    "DjangoRichHelpFormatter",
    "richify_command_line_help",
]


class DjangoRichHelpFormatter(_DjangoHelpFormatter, _RichHelpFormatter):
    """A rich help formatter for django commands."""


def richify_command_line_help(
    formatter_class: type[_RichHelpFormatter] = DjangoRichHelpFormatter,
) -> None:
    """Set a rich default formatter class for ``BaseCommand`` project-wide.

    Calling this function affects all built-in, third-party, and user defined django commands.

    Note that this function only changes the **default** formatter class of commands. User commands
    can still override the default by explicitly setting a formatter class.
    """
    from django.core.management.base import BaseCommand

    _patch_default_formatter_class(
        BaseCommand, formatter_class=formatter_class, method_name="create_parser"
    )
